{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Final Exam &mdash; Problem 6"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### SA421 Fall 2015"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**You will need to run the cell below for the basic constructs.**"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "import queue\n",
    "import collections\n",
    "import random\n",
    "import math\n",
    "\n",
    "\n",
    "# An event handler.\n",
    "class EventHandler:\n",
    "    def __init__(self):\n",
    "        self.time_now = 0\n",
    "        self.event_queue = queue.PriorityQueue()\n",
    "\n",
    "    def run(self, run_until_time=float('inf')):\n",
    "        while self.time_now <= run_until_time and (not self.event_queue.empty()):\n",
    "            # Get the next event from the queue\n",
    "            the_event = self.event_queue.get()\n",
    "            \n",
    "            if the_event.time_occurred > run_until_time:\n",
    "                break\n",
    "                \n",
    "            self.time_now = the_event.time_occurred\n",
    "            the_event.process(self)\n",
    "\n",
    "    def schedule(self, event):\n",
    "        self.event_queue.put(event)\n",
    "\n",
    "\n",
    "# A parent class for all random number generators\n",
    "class RandomNumberGenerator:\n",
    "    def next_random(self, handler: EventHandler):\n",
    "        return 0\n",
    "\n",
    "\n",
    "# A customer. Something who waits in queue to get served. Overide all methods for a special kind of customer.\n",
    "class Customer:\n",
    "    def __init__(self, name):\n",
    "        self.name = name\n",
    "        self.arrival_time = -1\n",
    "        self.start_service_time = -1\n",
    "        self.end_service_time = -1\n",
    "        self.departure_time = -1\n",
    "        self.my_server = None\n",
    "\n",
    "        self.my_servers = []\n",
    "        self.arrival_times = []\n",
    "        self.start_service_times = []\n",
    "        self.end_service_times = []\n",
    "\n",
    "        self.was_served = False\n",
    "        self.depart_on_service_done = True\n",
    "\n",
    "    def will_balk(self, service_queue):\n",
    "        return False\n",
    "\n",
    "    def handle_arrival(self, handler: EventHandler):\n",
    "        self.arrival_time = handler.time_now\n",
    "        self.arrival_times.append(handler.time_now)\n",
    "\n",
    "    def start_service(self, server, handler: EventHandler):\n",
    "        self.my_server = server\n",
    "        self.start_service_time = handler.time_now\n",
    "        self.start_service_times.append(handler.time_now)\n",
    "        return\n",
    "\n",
    "    def end_service(self, handler: EventHandler):\n",
    "        self.my_servers.append(self.my_server)\n",
    "        self.my_server = None\n",
    "        self.end_service_time = handler.time_now\n",
    "        self.end_service_times.append(handler.time_now)\n",
    "        self.was_served = True\n",
    "\n",
    "        # If I'm supposed to depart at the end of service, then I should do that now. Otherwise,\n",
    "        # this will not execute and I better o something else.\n",
    "        if self.depart_on_service_done:\n",
    "            handler.schedule(DepartureEvent(handler.time_now, self))\n",
    "\n",
    "    def handle_departure(self, handler: EventHandler):\n",
    "        self.my_server = None\n",
    "        self.departure_time = handler.time_now\n",
    "        # If we depart mid-service, fix up the times and servers.\n",
    "        if len(self.start_service_times) != len(self.end_service_times):\n",
    "            self.my_servers.append(self.my_server)\n",
    "            self.my_server = None\n",
    "            self.end_service_time = handler.time_now\n",
    "            self.end_service_times.append(handler.time_now)\n",
    "\n",
    "    def __str__(self):\n",
    "        return \"Customer[name=\" + self.name + \"]\"\n",
    "\n",
    "\n",
    "# A server who serves a customer.\n",
    "class Server:\n",
    "\n",
    "    def __init__(self):\n",
    "        self.now_serving = None\n",
    "        self.my_queue = None\n",
    "        self.service_time_rng = None\n",
    "        self.customers_served = []\n",
    "        self.start_service_times = []\n",
    "        self.end_service_times = []\n",
    "\n",
    "    def __init__(self, service_time_rng: RandomNumberGenerator):\n",
    "        self.now_serving = None\n",
    "        self.my_queue = None\n",
    "        self.service_time_rng = service_time_rng\n",
    "        self.customers_served = []\n",
    "        self.start_service_times = []\n",
    "        self.end_service_times = []\n",
    "\n",
    "    # Store the customer into the server's now_serving variable and generate the amount of time this\n",
    "    # service will require.\n",
    "    def start_service(self, customer, handler: EventHandler):\n",
    "        self.now_serving = customer\n",
    "        self.start_service_times.append(handler.time_now)\n",
    "        end_service_time = handler.time_now + self.compute_service_time(customer,handler)\n",
    "\n",
    "        # Schedule an event.\n",
    "        handler.schedule(ServiceDoneEvent(end_service_time, customer, self.my_queue))\n",
    "\n",
    "    def end_service(self, handler: EventHandler):\n",
    "        self.customers_served.append(self.now_serving)\n",
    "        self.end_service_times.append(handler.time_now)\n",
    "        self.now_serving = None\n",
    "\n",
    "    # You can override this to generate a custom service time computation.\n",
    "    def compute_service_time(self, customer, handler):\n",
    "        if self.service_time_rng is None:\n",
    "            return 0\n",
    "        else:\n",
    "            return self.service_time_rng.next_random(handler)\n",
    "\n",
    "    def compute_utilization(self, total_time):\n",
    "        sum = 0\n",
    "        num_services = len(self.customers_served)\n",
    "        for i in range(0,num_services):\n",
    "            dt = self.end_service_times[i] - self.start_service_times[i]\n",
    "            sum += dt\n",
    "\n",
    "        return sum / total_time\n",
    "\n",
    "\n",
    "# A Generic Event.\n",
    "class Event:\n",
    "    def __init__(self, time_occurred, event_type):\n",
    "        self.time_occurred = time_occurred\n",
    "        self.type = event_type\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        return\n",
    "\n",
    "    def __lt__(self, other):\n",
    "        if isinstance(other, Event):\n",
    "            return self.time_occurred < other.time_occurred\n",
    "        else:\n",
    "            # Default return true...in case you're forced into a queue with something weird\n",
    "            return True\n",
    "\n",
    "    def __str__(self):\n",
    "        return \"Event[type=\" + type + \", time_occurred=\" + str(self.time_occurred) + \"]\"\n",
    "\n",
    "\n",
    "# An event involving a customer, that may or may not be a queue event.\n",
    "class CustomerEvent(Event):\n",
    "    def __init__(self, time_occurred, customer: Customer, event_type):\n",
    "        Event.__init__(self, time_occurred, event_type)\n",
    "        self.customer = customer\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        return\n",
    "\n",
    "\n",
    "# A queue event. Something that happens to a queue.\n",
    "class QueueEvent(Event):\n",
    "    def __init__(self, time_occurred, customer, service_queue, event_type):\n",
    "        CustomerEvent.__init__(self, time_occurred, customer, event_type)\n",
    "        self.service_queue = service_queue\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        return\n",
    "\n",
    "\n",
    "# An event that models the arrival of some customer to a queue.\n",
    "class ArrivalEvent(QueueEvent):\n",
    "    def __init__(self, time_occurred, customer, service_queue):\n",
    "        QueueEvent.__init__(self, time_occurred, customer, service_queue, \"Arrival\")\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        self.service_queue.process_arrival(self.customer)\n",
    "\n",
    "    # Handle arrival events after departure, abandon and service done events.\n",
    "    def __lt__(self, other):\n",
    "        if isinstance(other, Event):\n",
    "            if self.time_occurred == other.time_occurred:\n",
    "                if isinstance(other, DepartureEvent) or isinstance(other, ServiceDoneEvent) or isinstance(other,\n",
    "                                                                                                          AbandonEvent):\n",
    "                    return False\n",
    "                else:\n",
    "                    return QueueEvent.__lt__(self, other)\n",
    "            else:\n",
    "                return QueueEvent.__lt__(self, other)\n",
    "        else:\n",
    "            return QueueEvent.__lt__(self, other)\n",
    "\n",
    "\n",
    "# An event that models the completion of service -- useful in queue networks, where departure doesn't necessarily happen\n",
    "# immediately.\n",
    "class ServiceDoneEvent(QueueEvent):\n",
    "    def __init__(self, time_occurred, customer, service_queue):\n",
    "        QueueEvent.__init__(self, time_occurred, customer, service_queue, \"ServiceDone\")\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        self.service_queue.process_service_done(self.customer)\n",
    "\n",
    "    # Handle before arrivals\n",
    "    def __lt__(self, other):\n",
    "        if isinstance(other, Event):\n",
    "            if self.time_occurred == other.time_occurred:\n",
    "                if isinstance(other, ArrivalEvent):\n",
    "                    return True\n",
    "                else:\n",
    "                    return QueueEvent.__lt__(self, other)\n",
    "            else:\n",
    "                return QueueEvent.__lt__(self, other)\n",
    "        else:\n",
    "            return QueueEvent.__lt__(self, other)\n",
    "\n",
    "\n",
    "# An event that models the departure of someone from the **system**.\n",
    "class DepartureEvent(CustomerEvent):\n",
    "    def __init__(self, time_occurred, customer):\n",
    "        CustomerEvent.__init__(self, time_occurred, customer, \"Departure\")\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        self.customer.handle_departure(handler)\n",
    "\n",
    "    # Handle before arrivals\n",
    "    def __lt__(self, other):\n",
    "        if isinstance(other, Event):\n",
    "            if self.time_occurred == other.time_occurred:\n",
    "                if isinstance(other, ArrivalEvent):\n",
    "                    return True\n",
    "                else:\n",
    "                    return QueueEvent.__lt__(self, other)\n",
    "            else:\n",
    "                return QueueEvent.__lt__(self, other)\n",
    "        else:\n",
    "            return QueueEvent.__lt__(self, other)\n",
    "\n",
    "\n",
    "# An event that models the abandonment of someone from a queue.\n",
    "class AbandonEvent(Event):\n",
    "    def __init__(self, time_occurred, customer, service_queue):\n",
    "        QueueEvent.__init__(self, time_occurred, customer, service_queue, \"Abandon\")\n",
    "        return\n",
    "\n",
    "    def process(self, handler: EventHandler):\n",
    "        self.service_queue.process_abandon(self.customer, self.time_occurred)\n",
    "\n",
    "    # Handle before arrivals\n",
    "    def __lt__(self, other):\n",
    "        if isinstance(other, Event):\n",
    "            if self.time_occurred == other.time_occurred:\n",
    "                if isinstance(other, ArrivalEvent):\n",
    "                    return True\n",
    "                else:\n",
    "                    return QueueEvent.__lt__(self, other)\n",
    "            else:\n",
    "                return QueueEvent.__lt__(self, other)\n",
    "        else:\n",
    "            return QueueEvent.__lt__(self, other)\n",
    "\n",
    "\n",
    "# A service queue (a resource in SimPy) this is something you wait in and ultimately get served in.\n",
    "class ServiceQueue:\n",
    "    def __init__(self, handler: EventHandler):\n",
    "        self.queue = collections.deque()\n",
    "        self.servers = []\n",
    "        self.event_handler = handler\n",
    "\n",
    "    def queue_length(self):\n",
    "        return len(self.queue)\n",
    "\n",
    "    def get_free_server(self):\n",
    "        available_servers = []\n",
    "        for server in self.servers:\n",
    "            if server.now_serving is None:\n",
    "                available_servers.append(server)\n",
    "\n",
    "        if len(available_servers) == 0:\n",
    "            return None\n",
    "        else:\n",
    "            return random.choice(available_servers)\n",
    "\n",
    "    def add_server(self, server):\n",
    "        server.my_queue = self\n",
    "        self.servers.append(server)\n",
    "\n",
    "    def call_next(self, server):\n",
    "        if len(self.queue) > 0:\n",
    "            next_customer = self.queue.pop()\n",
    "            next_customer.start_service(server, self.event_handler)\n",
    "            server.start_service(next_customer, self.event_handler)\n",
    "        return\n",
    "\n",
    "    def process_service_done(self, customer: Customer):\n",
    "        # Write code to process a departure from server\n",
    "        server = customer.my_server\n",
    "        leaving_customer = server.now_serving\n",
    "        leaving_customer.end_service(self.event_handler)\n",
    "        server.end_service(self.event_handler)\n",
    "        self.call_next(server)\n",
    "        return\n",
    "\n",
    "    def process_arrival(self, customer: Customer):\n",
    "        # Write code to process arrival\n",
    "        if customer.will_balk(self):\n",
    "            customer.handle_departure()\n",
    "            return\n",
    "\n",
    "        customer.handle_arrival(self.event_handler)\n",
    "        if self.queue_length() == 0:\n",
    "            available_server = self.get_free_server()\n",
    "            if available_server is None:\n",
    "                self.queue.appendleft(customer)\n",
    "            else:\n",
    "                self.queue.appendleft(customer)\n",
    "                self.call_next(available_server)\n",
    "        else:\n",
    "            self.queue.appendleft(customer)\n",
    "\n",
    "        return\n",
    "\n",
    "    # Allow customers to abandon if they're not in service.\n",
    "    def process_abandon(self, customer: Customer):\n",
    "        # Write code to process an abandonment\n",
    "        if self.queue.count(customer) > 0:\n",
    "            self.queue.remove(customer)\n",
    "            customer.handle_departure(self.event_handler)\n",
    "        return"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Part 1** Python's random number generator has mysteriously stopped working and you need some exponentially distributed random numbers. Fill in the code in <code>next_random()</code> to generate an exponentially distributed random variable. You will have to go through the process of generating a uniform random variable using a linear congruential generator. This essentially combines a number of things we've done into a single function call."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "\n",
    "class MyExpoRandomNumberGenerator(RandomNumberGenerator):\n",
    "    def __init__(self, mean_time_between_arrivals):\n",
    "        #Fill in this code.\n",
    "        \n",
    "        self.mean_time_between_arrivals = mean_time_between_arrivals\n",
    "        \n",
    "        #Multiplier\n",
    "        self.a = 1664525\n",
    "        \n",
    "        #Increment\n",
    "        self.C = 1013904223\n",
    "        \n",
    "        #Modulus\n",
    "        self.m = 2**32\n",
    "        \n",
    "        #Seed \n",
    "        self.X0 = 12345\n",
    "    \n",
    "    def next_random(self, handler: EventHandler):\n",
    "        #PART 1 -- Fill in the code to generate and store the next random integer\n",
    "        #Convert that integer into a uniform 0-1 value.\n",
    "        #Then generate an exponentially distributed random number\n",
    "        "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "**Part 2** Fill in the code so that a customer automatically generates the next customer on arrival. This is very similar to the automatically generating Sgt. Stan example but with variable names changed."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": [
    "class ExpoCustomer(Customer):\n",
    "    #This is a class variable. You can get at it by using ExpoCustomer.expo_customer_count. It's the *same* for every\n",
    "    #ExpoCustomer.\n",
    "    expo_customer_count = 0\n",
    "    \n",
    "    def __init__(self, name, arrival_time_rng: RandomNumberGenerator, service_queue: ServiceQueue):\n",
    "        Customer.__init__(self, name)\n",
    "        self.service_queue = service_queue\n",
    "        self.arrival_time_rng = arrival_time_rng\n",
    "\n",
    "    def schedule_next_arrival(self, handler: EventHandler):\n",
    "        #PART 2 -- Fill in the code to allow this customer to schedule the next arrival time for the next customer.\n",
    "\n",
    "\n",
    "    #Override the handle arrival event...now you can create the next vehicle in here.\n",
    "    def handle_arrival(self, handler: EventHandler):\n",
    "        Customer.handle_arrival(self, handler)\n",
    "        self.schedule_next_arrival(handler)\n",
    "\n",
    "\n",
    "def expo_simulation_auto_gen():\n",
    "    # This is the even handler.\n",
    "    handler = EventHandler()\n",
    "\n",
    "    # Create stan the server.\n",
    "    server = Server(MyExpoRandomNumberGenerator(15.0))\n",
    "\n",
    "    # Create the service queue. Put the server at the end of it.\n",
    "    service_queue = ServiceQueue(handler)\n",
    "    service_queue.add_server(server)\n",
    "\n",
    "    \n",
    "\n",
    "    expo_rng = MyExpoRandomNumberGenerator(20.0)\n",
    "    #You have to start the world somehow.\n",
    "    first_customer = ExpoCustomer(\"ExpoCustomer\", expo_rng, service_queue)\n",
    "    first_customer_time = expo_rng.next_random(handler)\n",
    "    handler.schedule(ArrivalEvent(first_customer_time, first_customer, service_queue))\n",
    "\n",
    "\n",
    "    handler.run(run_until_time=20000)\n",
    "    print(\"Server Utilitization = \" + str(server.compute_utilization(handler.time_now)))\n",
    "\n",
    "expo_simulation_auto_gen()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": true
   },
   "source": [
    "**Part 3** Run the code above and compare the server utilization after 20000 time units with the server utilization you expect to get from an M/M/1 queue."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "_Double click to type your answer here._"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true
   },
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.4.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 0
}
